/*
 * Sends SIGUSR1 to worker threads to expose races
 *
 * SIGSTOP would be natural, but it isn't per-thread.  So simulate it
 * via SIGUSR1 with a sleeping signal handler.  Also does cross-thread
 * allocation and free from the signal handler to trigger concurrency
 * bugs.
 */
#include <assert.h>
#include <pthread.h>
#include <signal.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/syscall.h>
#include <unistd.h>

/* These should become parameters */
#define default_iterations	(1024 * 1024 * 1024)
#define default_max_size	(64 * 1024)
#define default_max_mem		(1024 * 1024 * 1024)

#define iterations	(1024 * 1024)
#define max_size	(default_max_size)
#define max_mem		(default_max_mem)

#define thread_mem	(max_mem / no_threads)
#define array_size	(thread_mem / 1024)

#include "lib.c"

#define max_threads	(1024)
#define no_lists	(16 * max_threads)
static int no_threads;
void *cross_lists[no_lists];
pid_t thread_pid[max_threads];


static void cross_free(void *buf)
{
	static unsigned int global_index;
	unsigned int index = __sync_add_and_fetch(&global_index, 1) % no_lists;
	void *oldbuf;
	int i = 0;

	assert(index < no_lists);
	do {
		oldbuf = cross_lists[index];
		assert(i++ < 1000);
	} while (!__sync_bool_compare_and_swap(&cross_lists[index], oldbuf, buf));
	free(oldbuf);
}

static int running_threads;

static inline pid_t gettid(void)
{
	return syscall(SYS_gettid);
}

static void *thread_func(void *_seed)
{
	int seed = (long)_seed;
	struct mrand rdata;
	int i;
	void **array, **slot;
	size_t size, total_mem = array_size * sizeof(void*);

	thread_pid[seed] = gettid();
	mrand_init(&rdata, seed);
	assert(array = calloc(array_size, sizeof(void*)));
	for (i = iterations; i; i--) {
		slot = &array[mrand_max(&rdata, array_size - 1)];
		if (*slot) {
			total_mem -= check_free(*slot, slot);
			*slot = NULL;
		} else if (total_mem < thread_mem) {
			size = alloc_size(&rdata);
			total_mem += size;
			*slot = check_alloc(size, slot);
		}
	}
	thread_pid[seed] = 0;
	for (i = 0; i < array_size; i++) {
		slot = array[i];
		if (slot) {
			free(slot);
		}
	}
	free(array);
	__sync_sub_and_fetch(&running_threads, 1);
	return NULL;
}

void suspend_handler(int sig)
{
	cross_free(malloc(1));
	usleep(1000);
	cross_free(malloc(1));
}

static void main_thread(void)
{
	pthread_t thread[no_threads];
	struct sigaction sa;
	struct mrand rdata;
	int i;

	sigfillset(&sa.sa_mask);
	sa.sa_flags = 0;
	sa.sa_handler = suspend_handler;
	sigaction(SIGUSR1, &sa, NULL);

	mrand_init(&rdata, no_threads);
	for (i = 0; i < no_threads; i++) {
		__sync_add_and_fetch(&running_threads, 1);
		pthread_create(&thread[i], NULL, thread_func, (void *)(long)i);
	}
	while (running_threads) {
		int t = mrand_max(&rdata, no_threads - 1);
		if (!thread_pid[t]) {
			usleep(1000);
			continue;
		}
		kill(thread_pid[t], SIGUSR1);
		usleep(1000);
	}
	for (i = 0; i < no_threads; i++) {
		pthread_join(thread[i], NULL);
	}
}

int main(void)
{
	long cpus = sysconf(_SC_NPROCESSORS_ONLN);
	assert(cpus > 0);
	no_threads = cpus * 2;
	assert(no_threads <= max_threads);
	main_thread();
	return 0;
}
